import os
import shutil
import subprocess
import sys

class ExpressServer:

    def __init__(self):
        self.template = """
const express = require("express");
const bodyParser = require('body-parser');
const app = express();
app.use(bodyParser.json());
app.use(bodyParser.urlencoded({ extended: true }));
%s
%s
%s
app.listen(%s, () => console.log("Django Rest Mock Server: Listening on port %s"));""".strip()
        self.constructed = ''

    def __str__(self):
        if self.constructed:
            return self.constructed
        raise Exception('Call construct() first')

    def to_string(self):
        return str(self)

    def construct(self, variables, functions, endpoints, port=8000):
        self.constructed = self.template % (
            variables,
            functions,
            endpoints,
            port,
            port
        )

    def generate(self, file_path='index.js'):
        with open(file_path, 'w') as f:
            f.write(self.to_string())

    def start_server(self, file_path='index.js'):
        # Check if node is installed
        try:
            subprocess.check_output(['node', '-v'])
        except:
            sys.stdout.write('Node.js not installed properly.\n')
            sys.exit(0)

        # Check if Express is installed
        try:
            subprocess.check_output(['npm', 'list', 'express'])
        except:
            subprocess.call(['npm', 'install', 'express'])

        try:
            subprocess.call(["node", file_path])
        except KeyboardInterrupt:
            os.remove(file_path)
            path_prefix = os.getcwd()
            os.remove(path_prefix + '/package-lock.json')
            shutil.rmtree(path_prefix + '/node_modules')
            sys.stdout.write('Mock server shutdown.\n')



"""
Core functions to be placed in Express Server
"""

from .structures import Function


def data_finder():
    # If there are query params, first check if any of the queries match the pkName
    func_builder = Function()
    func_args = 'req'
    func_body = """
const path = req.path;
const method = req.method.toLowerCase();
const queries = req.query;
for (let i = 0; i < Object.keys(store).length; i ++) {
    if (Object.keys(store)[i].indexOf(path) > -1 && Object.keys(store)[i] === ('/' + method + '__' + path)) {
        const key = Object.keys(store)[i];
        const resp = store[key];
        const data = resp.data;
        return Object.assign({}, data);
    } else if (Object.keys(store)[i].indexOf('/' + method + '__' + path) > -1) {
        const key = Object.keys(store)[i];
        const resp = store[key];
        if (resp.pk && resp.position === 'query') {
            const keyName = resp.pkName;
            const query = req.query;
            let pk;
            if (Object.keys(query).length > 0 && query.constructor === Object) {
                pk = query[keyName];
            }
            if (pk && pk === resp.data[keyName]) {
                const data = resp.data;
                return Object.assign({}, data);
            }
        }
    }
}
return {};""".strip().replace('\n','')
    func_builder.construct('dataFinder', func_args, func_body)
    return str(func_builder.constructed) + ';'


def get_handler():
    """
    function getHandler (req) {
        const resp = dataFinder(req);
        return resp;
    }
    """
    func_builder = Function()
    func_args = 'req'
    func_body = "const resp = dataFinder(req);return resp;".strip().replace('\n','')
    func_builder.construct('getHandler', func_args, func_body)
    return str(func_builder.constructed) + ';'


def modify_handler():
    """
    function modifyHandler (req) {
        const path = req.path;
        const _newData = req.body;
        const newData = JSON.stringify(_newData);
        const storeCopy = Object.assign({}, store);
        const method = 'get';
        for (let i = 0; i < Object.keys(storeCopy).length; i ++) {
            if (Object.keys(storeCopy)[i].indexOf(path) > -1 && Object.keys(storeCopy)[i] === ('/' + method + '__' + path)) {
                const key = Object.keys(storeCopy)[i];
                if (req.method.toLowerCase() !== 'delete') {
                    storeCopy[key].data = Object.assign({}, storeCopy[key].data, newData);
                } else {
                    storeCopy[key].data = {};
                }
                store = Object.assign({}, store, storeCopy);
                const data = storeCopy[key].data;
                return data;
            } else if (Object.keys(storeCopy)[i].indexOf('/' + method + '__' + path) > -1) {
                const key = Object.keys(storeCopy)[i];
                const resp = storeCopy[key];
                if (resp.pk && resp.position === 'query') {
                    const keyName = resp.pkName;
                    const query = req.query;
                    let pk;
                    if (Object.keys(query).length > 0 && query.constructor === Object) {
                        pk = query[keyName];
                    }
                    if (pk && pk === resp.data[keyName]) {
                        if (req.method.toLowerCase() !== 'delete') {
                            storeCopy[key].data = Object.assign({}, storeCopy[key].data, newData);
                        } else {
                            storeCopy[key].data = {};
                        }
                        store = Object.assign({}, store, storeCopy);
                        const data = storeCopy[key].data;
                        return data;
                    }
                }
            }
        }
        return {};
    }
    """
    func_builder = Function()
    func_args = 'req'
    func_body = """
const path = req.path;
const newData = req.body;
const storeCopy = Object.assign({}, store);
const method = 'get';
for (let i = 0; i < Object.keys(storeCopy).length; i ++) {
    if (Object.keys(storeCopy)[i].indexOf(path) > -1 && Object.keys(storeCopy)[i] === ('/' + method + '__' + path)) {
        const key = Object.keys(storeCopy)[i];
        if (req.method.toLowerCase() !== 'delete') {
            storeCopy[key].data = Object.assign({}, storeCopy[key].data, newData);
        } else {
            storeCopy[key].data = {};
        }
        store = Object.assign({}, store, storeCopy);
        const data = storeCopy[key].data;
        return data;
    } else if (Object.keys(storeCopy)[i].indexOf('/' + method + '__' + path) > -1) {
        const key = Object.keys(storeCopy)[i];
        const resp = storeCopy[key];
        if (resp.pk && resp.position === 'query') {
            const keyName = resp.pkName;
            const query = req.query;
            let pk;
            if (Object.keys(query).length === 0 && query.constructor === Object) {
                pk = query[keyName];
            }
            if (pk && pk === resp.data.keyName) {
                if (req.method.toLowerCase() !== 'delete') {
                    storeCopy[key].data = Object.assign({}, storeCopy[key].data, newData);
                } else {
                    storeCopy[key].data = {};
                }
                store = Object.assign({}, store, storeCopy);
                const data = storeCopy[key].data;
                return data;
            }
        }
    }
}
return {};
    """.strip().replace('\n','')
    func_builder.construct('modifyHandler', func_args, func_body)
    return str(func_builder.constructed) + ';'


def post_handler():
    """
    function postHandler (req) {
        const resp = dataFinder(req);
        return resp;
    }
    """
    func_builder = Function()
    func_args = 'req'
    func_body = "const resp = dataFinder(req);return resp;".strip().replace('\n','')
    func_builder.construct('postHandler', func_args, func_body)
    return str(func_builder.constructed) + ';'

DATA_FINDER = data_finder()
GET_HANDLER = get_handler()
MODIFY_HANDLER = modify_handler()
POST_HANDLER = post_handler()



import ast
import json
import re
from copy import deepcopy
from urllib.parse import urlparse, parse_qs

from .core.express import ExpressServer
from .core.extractor import Extractor
from .core.functions import DATA_FINDER, GET_HANDLER, MODIFY_HANDLER, POST_HANDLER
from .core.parser import Parser
from .core.structures import Endpoint, Variable


def get_store(url_details):
    """
    The state of the mock server will be stored in the resulting
    object created here
    """
    store = {}
    for detail in url_details:
        base_url = detail['url'].strip()
        has_instances = False
        try:
            if len(detail['response']) > 1:
                for resp in detail['response']:
                    parsed_resp = ast.literal_eval(resp)
                    tmp_copy_resp = deepcopy(parsed_resp)
                    cleaned_resp = {}
                    # Remove meta keys from cleaned resp
                    for k, v in tmp_copy_resp.items():
                        if '__' not in k:
                            cleaned_resp[k] = v
                    if parsed_resp['__key_position'] == 'url':
                        unique_key = parsed_resp['__key_name']
                        constructed_url = re.sub(r'\_\_key', str(parsed_resp[unique_key]), base_url)
                        store[constructed_url] = {
                            'data': cleaned_resp,
                            'pk': True,
                            'pkName': parsed_resp['__key_name'],
                            'position': 'url',
                            'options': parsed_resp.get('__options', '{}')
                        }
                        has_instances = True
                    elif parsed_resp['__key_position'] == 'query':
                        unique_key = parsed_resp['__key_name']
                        if not base_url[-1] == '/':
                            url_with_slash = base_url + '/'
                        constructed_url = url_with_slash + '__pk/' + str(parsed_resp[unique_key])
                        constructed_url = re.sub(r'\_\_key', '', constructed_url)  # if '__key' hasn't already been replaced, remove it
                        store[constructed_url] = {
                            'data': cleaned_resp,
                            'pk': True,
                            'pkName': parsed_resp['__key_name'],
                            'position': 'query',
                            'options': parsed_resp.get('__options', '{}')
                        }
                        has_instances = True
                
                if has_instances:
                    # This endpoint has multiple instances
                    # we will need to create a new endpoint to list all
                    # these instances
                    list_url = re.sub(r'\/?\_\_key', '', base_url)
                    instances = []
                    for k, v in store.items():
                        if list_url in k:
                            instances.append(v['data'])
                    store[list_url] = {'data': instances}

            else:
                base_url = re.sub(r'\/?\_\_key', '', base_url)  # if '__key' hasn't already been replaced, remove it
                parsed_resp = ast.literal_eval(detail['response'][0])
                tmp_copy_resp = deepcopy(parsed_resp)
                cleaned_resp = {}
                # Remove meta keys from cleaned resp
                for k, v in tmp_copy_resp.items():
                    if '__' not in k:
                        cleaned_resp[k] = v

                store[base_url] = {
                    'data': cleaned_resp,
                    'pk': False,
                    'options': parsed_resp.get('__options', '{}')
                }
        except KeyError:
            # PUT, PATCH, DELETE
            pass
    return store


def clean_url(full_url, store, method):
    # Check in store if the base url has individual instances
    u = urlparse(full_url)
    url = '/' + method + '__' + u.path
    query = parse_qs(u.query)
    test_url = url.replace('__key', '(.*)')
    list_url = None
    r = re.compile(test_url)
    if '__key' in url:
        count = 0
        for k, v in store.items():
            options = v.get('options', '{}')
            exclude_key = False
            options = ast.literal_eval(options)
            if options is not None:
                exclude_key = options.get('excludeKey', False)
            search = r.search(k)
            if search is not None:
                list_url = re.sub(r'\/?\_\_key', '', url)
                count += 1
                if exclude_key:
                    url = re.sub(r'\/?\_\_key', '', url)
                    break
                url = re.sub(r'\/?\_\_key', '/:id', url)
                break
        if not count:
            url = re.sub(r'\/?\_\_key', '', url)
    elif query and '__key' in full_url:
        count = 0
        for k in store.keys():
            if r.search(k) is not None:
                list_url = url
                count += 1
                break
    return url, list_url


def build(port=8000):
    extractor = Extractor()
    parser = Parser(extractor.url_details)
    parser.parse()
    url_details = parser.results
    _store = get_store(url_details)
    store = json.dumps(_store)
    variables = str(Variable('let', 'store', store))
    functions = DATA_FINDER + GET_HANDLER + MODIFY_HANDLER + POST_HANDLER
    endpoints = []
    endpoint_uris = []
    for u in parser.results:
        endpoint = Endpoint()
        if u['method'].lower() in ['get', 'post']:
            method = u['method'].lower()
        else:
            method = 'modify'
        response = "const data = {}Handler(req);res.json(data);".format(method)

        # Check in store if the base url has individual instances
        u['url'], list_url = clean_url(u['full_url'], _store, u['method'].lower())
        if list_url is not None and u['method'].lower() == 'get':
            list_endpoint = Endpoint()
            list_endpoint.construct('get', list_url, response)
            if str(list_endpoint) not in endpoints:
                endpoints.append(str(list_endpoint))
            if list_endpoint.uri not in endpoint_uris:
                endpoint_uris.append(list_endpoint.uri)
        
        if method == 'modify':
            without_prefix = re.sub(r'\/(\w+)\_\_', '', u['url'])
            for k, v in _store.items():
                if without_prefix in k:
                    options = v.get('options', '{}')
                    options = ast.literal_eval(options)
                    modifiers = []
                    if options is not None:
                        modifiers = options.get('modifiers', [])
                    if modifiers:
                        for mod in modifiers:
                            if u['method'].lower() == mod:
                                mod_endpoint = Endpoint()
                                uri = without_prefix
                                if v.get('position') is not None and v['position'] == 'url':
                                    uri = re.sub(r'\/?\_\_key', '/:id', u['full_url'])
                                mod_endpoint.construct(u['method'].lower(), uri, response)
                                if str(mod_endpoint) not in endpoints:
                                    endpoints.append(str(mod_endpoint))
                                if mod_endpoint.uri not in endpoint_uris:
                                    endpoint_uris.append(mod_endpoint.uri)
        else:
            endpoint.construct(u['method'], u['url'], response)
            if str(endpoint) not in endpoints:
                endpoints.append(str(endpoint))
            if endpoint.uri not in endpoint_uris:
                endpoint_uris.append(endpoint.uri)

    endpoints = ''.join(endpoints)
    express = ExpressServer()
    express.construct(variables, functions, endpoints, port)
    return express



import os
from setuptools import setup, find_packages


here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst')).read()

setup(
    name='django-rest-mock',
    version='0.2.1',
    packages=find_packages(),
    description='Mock Express server generated based on your views, that can come in handy when developing REST APIs with Django',
    long_description=README,
    author='Thomas Jiang',
    author_email='thomasjiangcy@gmail.com',
    url='https://github.com/thomasjiangcy/django-rest-mock-server',
    license='MIT',
    install_requires=[
        'Django>=1.11',
        'Faker>=0.8.11'
    ],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        'Framework :: Django',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)



'''
Extracts all endpoints and their corresponding view docstrings to retrieve
the expected JSON responses.

Example
=======

Given a view::

    class SomeView(APIView):
        """
        URL: /api/some/view/__key?param1[int]&param2[str]
        """

        def get(self, request, *args, **kwargs):
            """
            ```
            {
                "__key": "<id:int>",
                "__key_position": "url",
                "__mockcount": 5,
                "data": "Hello, world",
                "id": "<sha256::10>"
            }
            ```
            """
            pass

The Extractor can parse the docstrings of the view class and the request handler method.
The query parameters and expected type should be parsed from the view class docstring.

'''

import re

import django
from django.conf import settings
from django.core.exceptions import ViewDoesNotExist


if django.VERSION >= (2, 0):
    from django.urls import URLPattern, URLResolver

    class RegexURLPattern:
        pass

    class RegexURLResolver:
        pass

    class LocaleRegexURLResolver:
        pass

    def describe_pattern(p):
        return str(p.pattern)
else:
    try:
        from django.urls import RegexURLPattern, RegexURLResolver, LocaleRegexURLResolver
    except ImportError:
        from django.core.urlresolvers import RegexURLPattern, RegexURLResolver, LocaleRegexURLResolver

    class URLPattern:
        pass

    class URLResolver:
        pass

    def describe_pattern(p):
        return p.regex.pattern


METHODS = ['get', 'post', 'put', 'patch', 'delete']


class Extractor:

    def __init__(self):
        self._urlconf = __import__(getattr(settings, 'ROOT_URLCONF'))
        self.url_details = []
        self._load_url_details()  # Load all urlpatterns & corresponding view classes/functions

    def _get_view_details(self, urlpatterns, parent=''):
        """Recursive function to extract all url details"""
        for pattern in urlpatterns:
            if isinstance(pattern, (URLPattern, RegexURLPattern)):
                try:
                    d = describe_pattern(pattern)
                    docstr = pattern.callback.__doc__
                    method = None
                    expected_json_response = ''
                    
                    expected_url = ''
                    if docstr:
                        # Get expected URL
                        u = re.findall(r'URL: (.*)', docstr, flags=re.DOTALL)
                        if u:
                            expected_url = u[0]

                    # Get all possible methods
                    if 'view_class' not in dir(pattern.callback):
                        continue

                    possible_methods = [m for m in dir(pattern.callback.view_class) if m in METHODS]

                    for method in possible_methods:
                        view_method_docstr = getattr(pattern.callback.view_class, method).__doc__
                        if view_method_docstr:
                            # Check if method is modifier
                            if method in ['put', 'patch', 'delete']:
                                self.url_details.append({
                                    'url': expected_url,
                                    'method': method
                                })
                                continue
                            # Extract request method and JSON response from docstring of request method
                            j = re.findall(r'```(.*)```', view_method_docstr, flags=re.DOTALL)
                            if j is not None:
                                for match in j:
                                    expected_json_response = match.strip()
                                    # Only add the details if all 3 values are filled
                                    if method is not None and expected_json_response and expected_url:
                                        self.url_details.append({
                                            'url': expected_url,
                                            'method': method,
                                            'response': expected_json_response
                                        })
                                        continue

                except ViewDoesNotExist:
                    pass

            elif isinstance(pattern, (URLResolver, RegexURLResolver)):
                try:
                    patterns = pattern.url_patterns
                except ImportError:
                    continue

                d = describe_pattern(pattern)
                current_full_url = parent + d
                self._get_view_details(patterns, current_full_url)
    
    def _load_url_details(self):
        root_urls = self._urlconf.urls.urlpatterns
        self._get_view_details(root_urls)



"""
Parser needs to make sense of:

* Query params
* Request method - whether it is a retrieval or modification (GET vs the rest)
* Data to generate (and the corresponding number of endpoints) - uses Faker to generate

"""

import ast
import json
import re
from copy import deepcopy
from urllib.parse import urlparse, parse_qs

from faker import Faker


PYTHON_DATATYPES = ['str', 'int', 'float', 'decimal', 'set', 'dict', 'list', 'tuple', 'bool']
fake = Faker()


def validate_fake_val(fake_val, fake_func, minimum, maximum):
    if (type(fake_val) == 'int' or type(fake_val) == 'float' or type(fake_val) == 'decimal') and\
        (minimum or maximum):
        if minimum and not maximum:
            while fake_val < minimum:
                fake_val = fake_func()  # keep trying until fake_val > minimum
        elif minimum and maximum:
            while fake_val < minimum or fake_val > maximum:
                fake_val = fake_func()  # keep trying until minimum <= fake_val <= maximum
        elif not minimum and maximum:
            while fake_val > maximum:
                fake_val = fake_func()  # keep trying until fake_val <= maximum

    elif type(fake_val) == 'str' and (minimum or maximum):
        if minimum and not maximum:
            fake_val = fake_val[minimum:]
        elif minimum and maximum:
            fake_val = fake_val[minimum:maximum]
        elif not minimum and maximum:
            fake_val = fake_val[:maximum]

    # For all other data types, min/max attributes will have no effect
    return str(fake_val)  # make sure to return a 'str' type


def replace_faker_attr(matchobj):
    attr = matchobj.group(0).replace('<', '').replace('>', '')
    # Check if optional arguments are specified - syntax is <attr:min:max>
    _split = attr.split(":")
    if len(_split) > 1:
        # Get the attr and value range (min/max)
        attr, val_range = _split[0], _split[1:]
        if len(val_range) > 1:
            minimum, maximum = val_range
        else:
            minimum, maximum = val_range[0], 0
    else:
        attr, minimum, maximum = _split[0], 0, 0
    # Min, max will be applied in a slice function if the data is a string
    if attr in PYTHON_DATATYPES:
        attr = 'py' + attr

    try:
        fake_func = getattr(fake, attr)
        fake_val = fake_func()
        fake_val = validate_fake_val(fake_val, fake_func, minimum, maximum)
        return fake_val
    except AttributeError:
        return attr


class Parser:
    
    def __init__(self, url_details):
        self.url_details = url_details
        self._results = None
    
    @property
    def results(self):
        if self._results is not None:
            return self._results
        raise Exception('URL details need to be parsed first, call parse() to generate results.')
    
    @staticmethod
    def _parse_url(url_string):
        u = urlparse(url_string)
        url = u.path
        query = parse_qs(u.query)

        return {
            'full_url': url_string.strip(),
            'url': url.strip(),
            'query': query
        }

    @staticmethod
    def _parse_response(response):
        resp = json.loads(response)
        count = resp.get("__mockcount", 1)
        generated_responses = []
        for _ in range(count):
            response = {}
            key_type, key_name, key_position = [None] * 3
            for k, v in resp.items():
                if k == '__mockcount' or k == '__key_position':
                    continue
                if k == '__key':
                    try:
                        key_position = resp['__key_position']
                    except KeyError:
                        raise Exception("'__key_position' must be present if '__key' is present")
                    v = v.replace('<', '').replace('>', '')
                    key_name, key_type = v.split(':')
                    
                # Check if value fulfils the dynamic syntax
                v_type = type(deepcopy(v)).__name__
                v = str(v)  # treat the value as a string regardless of its actual data type
                has_syntax = re.findall(r'<\w+(\d+)?(\:)?(\d+)?(\:)?(\d+)?>', v, flags=re.DOTALL)

                if has_syntax:
                    fake_val = re.sub(r'<\w+(\d+)?(\:)?(\d+)?(\:)?(\d+)?>', replace_faker_attr, v, flags=re.DOTALL)
                    if v_type in ['list', 'dict', 'tuple', 'set']:
                        response[k] = ast.literal_eval(fake_val)
                    else:
                        response[k] = fake_val
                else:
                    if k == '__key': continue
                    response[k] = v

            response['__key_type'] = key_type
            response['__key_position'] = key_position
            response['__key_name'] = key_name
            generated_responses.append(str(response))
        return generated_responses
    
    def parse(self):
        results = []
        for url in self.url_details:
            url_detail = self._parse_url(url['url'])
            method = 'get' if url['method'].lower() in ['get', 'put', 'patch', 'delete'] else 'post'
            url_detail['url'] = '/' + method + '__' + url_detail['url']
            url_detail['method'] = url['method']
            try:
                url_detail['response'] = self._parse_response(url['response'])
            except KeyError:
                pass
            results.append(url_detail)
        self._results = results



from django.core.management.base import BaseCommand

from rest_mock_server.builder import build


class Command(BaseCommand):
    help = 'Generates an Express server file for mocking endpoint responses'

    def add_arguments(self, parser):
        parser.add_argument(
            '--output',
            dest='output_file',
            help='Custom output file name'
        )
        parser.add_argument(
            '--port',
            dest='port',
            help='Custom port to be exposed'
        )

    def handle(self, *args, **options):
        output = options.get('output_file')
        if output is None:
            output = 'index.js'
        port = options.get('port')
        if port is None:
            port = 8000
        express = build(port)
        express.generate(file_path=output)



import re

class BaseFactory:

    def __init__(self):
        self.constructed = ''
        self.options = dict(strip=True, remove_space=True, strip_newline=False, strip_tab=False)
        self.is_constructed = False

    def __str__(self):
        if self.is_constructed:
            return self.constructed
        else:
            raise Exception('Need to call construct() first')


class Variable(BaseFactory):

    def __init__(self, var_type, var_name, var_val):
        super().__init__()
        self.constructed = "%s %s = %s;"
        self.construct(var_type, var_name, var_val)
    
    def construct(self, var_type, var_name, var_val):
        self.constructed = self.constructed % (
            var_type,
            var_name,
            var_val
        )
        self.is_constructed = True


class Function(BaseFactory):

    def __init__(self):
        super().__init__()
        self.constructed = "function %s(%s) {%s};"

    def construct(self, name, args, body):
        self.constructed = self.constructed % (
            name,
            args,
            body
        )
        self.is_constructed = True


class Endpoint(BaseFactory):

    def __init__(self):
        super().__init__()
        self.constructed = 'app.%s("%s", (req, res) => {%s});\n'
        self.uri = None

    def construct(self, method, uri, response):
        cleaned_uri = re.findall(r'\/\w+\_\_(\/.*)', uri)
        if cleaned_uri:
            uri = cleaned_uri[0]
        self.constructed = self.constructed % (method, uri, response)
        self.uri = uri
        self.is_constructed = True
